import clr
import sys
import csv
import copy
import math
import os
import System

from timeit import default_timer as timer
from datetime import timedelta

clr.AddReference ('IronPython.Wpf')
import wpf

from System import Math, Action, Func, String
from System.Text import Encoding
from System.Collections.Generic import List, IEnumerable
from System.Windows import Visibility
from System.Windows.Controls import StackPanel, ComboBox, ComboBoxItem, ListBox, ListBoxItem, ItemCollection
from System.Windows.Input import Keyboard, AccessKeyManager
from System import Type, Tuple, Array, Double
from System.IO import StreamReader, MemoryStream, File
from System.Windows.Threading import Dispatcher
clr.AddReference ("System.Drawing")
from System.Drawing import Bitmap, Color
clr.AddReference ("System.Windows.Forms")
from System.Windows.Forms import Control, Keys, OpenFileDialog, DialogResult, FolderBrowserDialog, ColumnHeader 

# we are referencing way too many Trimble functions
# but that way it is easier to fix in case they move stuff around again
try:
    clr.AddReference ("Trimble.Sdk") # In version 5.50, model assemblies are "pre-referenced"
    clr.AddReference ("Trimble.Sdk.Primitives") # contains Trimble.Vce.Collections
    clr.AddReference ("Trimble.Sdk.GraphicsModel") # Trimble.Vce.TTMetrics

except:
    clr.AddReference ("Trimble.DiskIO")
    clr.AddReference ("Trimble.Vce.Alignment")
    clr.AddReference ("Trimble.Vce.Collections")
    clr.AddReference ("Trimble.Vce.Coordinates")
    clr.AddReference ("Trimble.Vce.Core")
    clr.AddReference ("Trimble.Vce.Data")
    clr.AddReference ("Trimble.Vce.Data.COGO")
    clr.AddReference ("Trimble.Vce.Data.Construction")
    clr.AddReference ("Trimble.Vce.Data.Scanning")
    clr.AddReference ("Trimble.Vce.Data.RawData")
    clr.AddReference ("Trimble.Vce.ForeignCad")
    clr.AddReference ("Trimble.Vce.Gem")
    clr.AddReference ("Trimble.Vce.Geometry")
    clr.AddReference ("Trimble.Vce.Interfaces")
    clr.AddReference ("Trimble.Vce.TTMetrics")
    clr.AddReference ("Trimble.Vce.Units")
    clr.AddReference ("Trimble.Vce.Utility")

from Trimble.DiskIO import OptionsManager

from Trimble.Vce.Alignment import ProfileView, VerticalAlignment, Linestring, AlignmentLabel, DrapedLine

from Trimble.Vce.Alignment.Linestring import Linestring,ElementFactory, IStraightSegment, IXYZLocation, IPointIdLocation

from Trimble.Vce.Collections import Point3DArray, DynArray

from Trimble.Vce.Coordinates import Point as CoordPoint, ICoordinate, PointCollection

from Trimble.Vce.Core import TransactMethodCall, ModelEvents

from Trimble.Vce.Core.Components import WorldView, Project, Layer, LayerGroupCollection, LayerCollection, TextStyle, TextStyleCollection, LineStyleCollection, BlockView, EntityContainerBase, PointLabelStyle, SnapInAttributeExtension

from Trimble.Vce.Data import PointCloudSelection, FileProperties, FilePropertiesContainer

from Trimble.Vce.Data.COGO import Calc

from Trimble.Vce.Data.Construction.IFC import IFCMesh

from Trimble.Vce.Data.Construction.Settings import ConstructionCommandsSettings

from Trimble.Vce.Data.RawData import PointManager, RawDataContainer

clr.AddReference ("Trimble.Vce.Data.RXL")
from Trimble.Vce.Data.RXL import RXLAlignmentExporter, FileWriter as RxlFileWriter, Versions as RxlVersions

from Trimble.Vce.Data.Scanning import PointCloudRegion, ExposedPointCloudRegion

from Trimble.Vce.ForeignCad import PolyLine, Poly3D, Arc as ArcObject, Circle as CadCircle # we are also using Arc from geometry
from Trimble.Vce.ForeignCad import Point as CadPoint, MText, Text as CadText, BlockReference, Leader, LeaderType, DimArrowheadType, AttachmentPoint

from Trimble.Vce.Gem import VextexAndTriangleList, Model3D, ProjectedSurface, Gem, Filer, Model3DCompSettings, IModel3D

from Trimble.Vce.Geometry import Triangle2D, Triangle3D, Point3D, Plane3D, Arc, Matrix4D, Vector2D, Vector3D, BiVector3D, Spinor3D, PolySeg, Limits3D, Side, Intersections, Rectangle3D, RectangleD, Primitive, PrimitiveLocation, Conversions

from Trimble.Vce.Geometry.PolySeg.Segment import Segment, Line as SegmentLine, Arc as ArcSegment, Parabola as ParabolaSegment

from Trimble.Vce.Interfaces import Client, ProgressBar

from Trimble.Vce.Interfaces.Client import CommandGranularity

from Trimble.Vce.Interfaces.Core import OverrideByLayer

from Trimble.Vce.Interfaces.Construction import UtilityNodeType

from Trimble.Vce.Interfaces.SnapIn import IPolyseg, IName, IHaveUcs, ISnapIn, IPointCloudReader, IMemberManagement, ICollectionOfEntities, TransformData

from Trimble.Vce.Interfaces.PointCloud import IExposedPointCloudRegion

from Trimble.Vce.SurveyCAD import ProjectLineStyle

from Trimble.Vce.TTMetrics import StrokeFont, StrokeFontManager

from Trimble.Vce.Utility import UtilityNode

clr.AddReference ("Trimble.Vce.UI.BaseCommands")
from Trimble.Vce.UI.BaseCommands import ViewHelper, SelectionContextMenuHandler, ExploreObjectControlHelper

clr.AddReference ("Trimble.Vce.UI.ConstructionCommands")
from Trimble.Vce.UI.ConstructionCommands import LimitSliceWindow

clr.AddReference ("Trimble.Vce.UI.Controls")
from Trimble.Vce.UI.Controls import SurfaceTypeLists, GlobalSelection, TrimbleColor, ExplorerUI, ExplorerItemCollection

clr.AddReference("Trimble.Vce.UI.Hoops")
from Trimble.Vce.UI.Hoops import Hoops2dView, Hoops3dView

clr.AddReference("Trimble.Vce.UI.UIManager")
from Trimble.Vce.UI.UIManager import UIEvents, TrimbleOffice

# later version of TBC moved these to a different assembly. If not found in new location, look at old
try:
	#5.50
    from Trimble.Sdk.Interfaces.UI import InputMethod, MousePosition, CursorStyle, ControlBoolean, UIEventArgs, I2DProjection, DistanceType
except:
    try:
		#5.40 or so
        from Trimble.Vce.UI.UIManager import UIEventArgs
        from Trimble.CustomControl.Interfaces import MousePosition, ControlBoolean, I2DProjection
        from Trimble.CustomControl.Interfaces.Enums import CursorStyle, InputMethod
    except:
		# even older
        from Trimble.Vce.UI.Controls import MousePosition, CursorStyle, ControlBoolean, TrimbleColor


def Setup(cmdData, macroFileFolder):
    cmdData.Key = "SCR_ManholeSurfaceRL"
    cmdData.CommandName = "SCR_ManholeSurfaceRL"
    cmdData.Caption = "_SCR_ManholeSurfaceRL"
    cmdData.UIForm = "SCR_ManholeSurfaceRL"      # MUST MATCH NAME FROM CLASS DEFINED BELOW !!!
    cmdData.HelpFile = "Macros.chm"
    cmdData.HelpTopic = "22602"

    try:
        cmdData.DefaultTabKey = "SCR ImExport/DTM/Subgrade"
        cmdData.DefaultTabGroupKey = "DTM"
        cmdData.ShortCaption = "Manhole RL"
        cmdData.DefaultRibbonToolSize = 3 # Default=0, ImageOnly=1, Normal=2, Large=3

        cmdData.Version = 1.10
        cmdData.MacroAuthor = "SCR"
        cmdData.MacroInfo = r""
        
        cmdData.ToolTipTitle = "set Manhole-RL to Surface-RL"
        cmdData.ToolTipTextFormatted = "change Utility-Manhole-RLs to Surface-RL"

    except:
        pass
    try:
        b = Bitmap (macroFileFolder + "\\" + cmdData.Key + ".png") 
        cmdData.ImageSmall = b
    except:
        pass

class SCR_ManholeSurfaceRL(StackPanel): # this inherits from the WPF StackPanel control
    def __init__(self, currentProject, macroFileFolder):
        with StreamReader (macroFileFolder + r"\SCR_ManholeSurfaceRL.xaml") as s:
            wpf.LoadComponent (self, s)
        self.currentProject = currentProject


    def OnLoad(self, cmd, buttons, event):
        
        self.okBtn = buttons[0]
        self.Caption = cmd.Command.Caption

        types = Array [Type] (SurfaceTypeLists.AllWithCutFillMap)    # we fill an array with TBC object types, we could combine different types
        self.surfacepicker.FilterByEntityTypes = types    # we fill the dropdownlist by applying that types array as filter
        self.surfacepicker.AllowNone = False              # our list shall not show an empty field

        # get the units for linear distance
        self.lunits = self.currentProject.Units.Linear
        self.lfp = self.lunits.Properties.Copy()
        linearsuffix = self.lunits.Units[self.lunits.DisplayType].Abbreviation
        #self.lfp.AddSuffix = False
        self.verticaloffsetlabel.Content = "Vertical Offset [" + linearsuffix + "]"
        self.horizontaloffsetlabel.Content = "Horizontal Offset [" + linearsuffix + "]"

        self.verticaloffset.NumberOfDecimals = self.lunits.Units[self.lunits.DisplayType].NumberOfDecimals
        self.horizontaloffset.NumberOfDecimals = self.lunits.Units[self.lunits.DisplayType].NumberOfDecimals

        self.objs.IsEntityValidCallback=self.IsValid

        self.SetDefaultOptions()

    def IsValid(self, serial):
        o = self.currentProject.Concordance.Lookup(serial)
        if isinstance(o, UtilityNode):
            return True
        return False
        
    def SetDefaultOptions(self):
        try:    self.surfacepicker.SelectIndex(OptionsManager.GetInt("SCR_ManholeSurfaceRL.surfacepicker", 0))
        except: self.surfacepicker.SelectIndex(0)
        self.adjustrim.IsChecked = OptionsManager.GetBool("SCR_ManholeSurfaceRL.adjustrim", True)
        self.adjustfloor.IsChecked = OptionsManager.GetBool("SCR_ManholeSurfaceRL.adjustfloor", False)
        self.verticaloffset.Value = OptionsManager.GetDouble("SCR_ManholeSurfaceRL.verticaloffset", 0.0)
        self.createlines.IsChecked = OptionsManager.GetBool("SCR_ManholeSurfaceRL.createlines", False)
        self.horizontaloffset.Value = OptionsManager.GetDouble("SCR_ManholeSurfaceRL.horizontaloffset", 0.0)
        
        settingserial = OptionsManager.GetUint("SCR_ManholeSurfaceRL.layerpicker", 8) # 8 is FixedSerial for Layer Zero
        o = self.currentProject.Concordance.Lookup(settingserial)
        if o != None:
            if isinstance(o.GetSite(), LayerCollection):    
                self.layerpicker.SetSelectedSerialNumber(settingserial, InputMethod(3))
            else:                       
                self.layerpicker.SetSelectedSerialNumber(8, InputMethod(3))
        else:                       
            self.layerpicker.SetSelectedSerialNumber(8, InputMethod(3))


    def SaveOptions(self):
        OptionsManager.SetValue("SCR_ManholeSurfaceRL.adjustrim", self.adjustrim.IsChecked)
        OptionsManager.SetValue("SCR_ManholeSurfaceRL.adjustfloor", self.adjustfloor.IsChecked)
        OptionsManager.SetValue("SCR_ManholeSurfaceRL.surfacepicker", self.surfacepicker.SelectedIndex)
        OptionsManager.SetValue("SCR_ManholeSurfaceRL.verticaloffset", self.verticaloffset.Value)
        OptionsManager.SetValue("SCR_ManholeSurfaceRL.createlines", self.createlines.IsChecked)
        OptionsManager.SetValue("SCR_ManholeSurfaceRL.horizontaloffset", self.horizontaloffset.Value)
        OptionsManager.SetValue("SCR_ManholeSurfaceRL.layerpicker", self.layerpicker.SelectedSerialNumber)

    def createlinesChanged(self, sender, e):
        if self.createlines.IsChecked:
            self.surfacepicker.IsEnabled = False
            self.verticaloffset.IsEnabled = False
            self.horizontaloffset.IsEnabled = True
            self.layerpicker.IsEnabled = True
        else:
            self.surfacepicker.IsEnabled = True
            self.verticaloffset.IsEnabled = True
            self.horizontaloffset.IsEnabled = False
            self.layerpicker.IsEnabled = False

    def OkClicked(self, cmd, e):
        Keyboard.Focus(self.okBtn)
        self.error.Content=''
        self.success.Content=''


        self.currentProject.TransactionManager.AddBeginMark(CommandGranularity.Command, self.Caption)
        UIEvents.RaiseBeforeDataProcessing(self, UIEventArgs())

        wv = self.currentProject [Project.FixedSerial.WorldView]

        surface = wv.Lookup(self.surfacepicker.SelectedSerial)

        try:
            # the "with" statement will unroll any changes if something go wrong
            with TransactMethodCall(self.currentProject.TransactionCollector) as failGuard:

                # TBC internally computes in meter
                verticaloffset_meter = self.lunits.Convert(self.lunits.DisplayType, self.verticaloffset.Value, self.lunits.InternalType)

                if self.objs.Count == 0:
                    self.error.Content += '\nnothing selected'

                for o in self.objs:
                    if isinstance(o, UtilityNode):
                        
                        if not self.createlines.IsChecked: # compute DTM elevation and change node values
                            p = o.Point0

                            surfp = surface.PickSurface(p)
                            if surfp[0]:

                                if self.adjustrim.IsChecked:

                                    o.RimElevation = surfp[1].Z + verticaloffset_meter
                
                                else:

                                    p.Z = surfp[1].Z + verticaloffset_meter
                                    o.Point0 = p

                            else:
                                self.error.Content += "\ncouldn't compute a valid surface elevation"

                        else: # create outlines

                            # get the elevation
                            if self.adjustrim.IsChecked: 
                                el = o.RimElevation
                            else:
                                el = o.FloorElevation

                            horizontaloffset_meter = self.lunits.Convert(self.lunits.DisplayType, self.horizontaloffset.Value, self.lunits.InternalType)

                            # Manhole = 6; Eccentric Manhole = 7; Junctionbox = 12
                            if o.NodeType == UtilityNodeType(6) or o.NodeType == UtilityNodeType(7):
                                cp = o.Point0
                                cp.Z = el
                                
                                circle = wv.Add(clr.GetClrType(CadCircle))
                                circle.CenterPoint = cp
                                circle.Radius = (o.SiteImprovement.Diameter / 2) + o.SiteImprovement.Thickness
                                circle.Layer = self.layerpicker.SelectedSerialNumber

                                if horizontaloffset_meter <> 0.0:
                                    os_circle = wv.Add(clr.GetClrType(CadCircle))
                                    os_circle.CenterPoint = cp
                                    os_circle.Radius = (o.SiteImprovement.Diameter / 2) + \
                                                        o.SiteImprovement.Thickness + horizontaloffset_meter
                                    os_circle.Layer = self.layerpicker.SelectedSerialNumber

                            elif o.NodeType == UtilityNodeType(12):
                                boxlength = o.SiteImprovement.Length + (2 * o.SiteImprovement.Thickness)
                                boxwidth  = o.SiteImprovement.Width + (2 * o.SiteImprovement.Thickness)

                                boxvec = Vector3D(Vector3D(o.Point0), 1.0)
                                boxvec.Azimuth = 2 * math.pi - o.Orientation + math.pi/2
                               
                                l = wv.Add(clr.GetClrType(Linestring))
                                l.Layer = self.layerpicker.SelectedSerialNumber
                                l.Closed = True

                                # get to the first corner, adding vectors and rotating them
                                tmpp = o.Point0
                                tmpp.Z = el
                                boxvec.Length = boxlength / 2
                                tmpp += boxvec
                                boxvec.Rotate90(Side.Left)
                                boxvec.Length = boxwidth / 2
                                tmpp += boxvec

                                e = ElementFactory.Create(clr.GetClrType(IStraightSegment), clr.GetClrType(IXYZLocation))
                                e.Position = tmpp
                                l.AppendElement(e)
                                
                                # go to the 2nd corner
                                boxvec.Rotate90(Side.Left)
                                boxvec.Length = boxlength
                                tmpp += boxvec

                                e = ElementFactory.Create(clr.GetClrType(IStraightSegment), clr.GetClrType(IXYZLocation))
                                e.Position = tmpp
                                l.AppendElement(e) 

                                # go to the 3rd corner
                                boxvec.Rotate90(Side.Left)
                                boxvec.Length = boxwidth
                                tmpp += boxvec

                                e = ElementFactory.Create(clr.GetClrType(IStraightSegment), clr.GetClrType(IXYZLocation))
                                e.Position = tmpp
                                l.AppendElement(e) 

                                # go to the 4th corner
                                boxvec.Rotate90(Side.Left)
                                boxvec.Length = boxlength
                                tmpp += boxvec

                                e = ElementFactory.Create(clr.GetClrType(IStraightSegment), clr.GetClrType(IXYZLocation))
                                e.Position = tmpp
                                l.AppendElement(e) 

                                if horizontaloffset_meter <> 0.0:
                                    # computing and creating the offset line
                                    l_h = l.ComputePolySeg()
                                    l_v = l.ComputeVerticalPolySeg()
                                    poly_offset = l_h.Offset(Side.Right, horizontaloffset_meter)
                                    
                                    if poly_offset[0] == True: # if the offset was computed
                                        l = wv.Add(clr.GetClrType(Linestring))
                                        l.Layer = self.layerpicker.SelectedSerialNumber
                                        l.Append(poly_offset[1], l_v, False, False)
                            else:
                                self.error.Content += '\nfound an unsupported NodeType for line creation'
                
                failGuard.Commit()
                UIEvents.RaiseAfterDataProcessing(self, UIEventArgs())
                self.currentProject.TransactionManager.AddEndMark(CommandGranularity.Command)
        
        except Exception as e:
            tt = sys.exc_info()
            exc_type, exc_obj, exc_tb = sys.exc_info()
            # EndMark MUST be set no matter what
            # otherwise TBC won't work anymore and needs to be restarted
            self.currentProject.TransactionManager.AddEndMark(CommandGranularity.Command)
            UIEvents.RaiseAfterDataProcessing(self, UIEventArgs())
            self.error.Content += '\nan Error occurred - Result probably incomplete\n' + str(exc_type) + '\n' + str(exc_obj) + '\nLine ' + str(exc_tb.tb_lineno)

        self.success.Content += '\Done'
        self.SaveOptions()  